"""
The :mod:`imblearn.model_selection` provides methods to split the dataset into
training and test sets.
"""

from imblearn.model_selection._split import InstanceHardnessCV

__all__ = ["InstanceHardnessCV"]
