@import url("theme.css");

.highlight a {
  text-decoration: underline;
}

.deprecated p {
  padding: 10px 7px 10px 10px;
  color: #b94a48;
  background-color: #f3e5e5;
  border: 1px solid #eed3d7;
}

.deprecated p span.versionmodified {
  font-weight: bold;
}

.wy-nav-content {
  max-width: 1200px !important;
}

/* Override some aspects of the pydata-sphinx-theme */

/* Main index page overview cards */

.intro-card {
  padding: 30px 10px 20px 10px;
}

.intro-card .sd-card-img-top {
  margin: 10px;
  height: 52px;
  background: none !important;
}

.intro-card .sd-card-title {
  color: var(--pst-color-primary);
  font-size: var(--pst-font-size-h5);
  padding: 1rem 0rem 0.5rem 0rem;
}

.intro-card .sd-card-footer {
  border: none !important;
}

.intro-card .sd-card-footer p.sd-card-text {
  max-width: 220px;
  margin-left: auto;
  margin-right: auto;
}

.intro-card .sd-btn-secondary {
  background-color: #6c757d !important;
  border-color: #6c757d !important;
}

.intro-card .sd-btn-secondary:hover {
  background-color: #5a6268 !important;
  border-color: #545b62 !important;
}

.card, .card img {
  background-color: var(--pst-color-background);
}
